/**
 * Copyright Benjamin Piwowarski, 2010. All rights reserved. Permission to use,
 * copy, modify and redistribute is granted, provided that this copyright
 * notice is retained and the author is given credit whenever appropriate.
 * 
 * Based on ArgParser:
 * Copyright John E. Lloyd, 2004. All rights reserved. Permission to use,
 * copy, modify and redistribute is granted, provided that this copyright
 * notice is retained and the author is given credit whenever appropriate.
 *
 * This  software is distributed "as is", without any warranty, including 
 * any implied warranty of merchantability or fitness for a particular
 * use. The author assumes no responsibility for, and shall not be liable
 * for, any special, indirect, or consequential damages, or any damages
 * whatsoever, arising out of or in connection with the use of this
 * software.
 */
package bpiwowar.argparser;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.io.Reader;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.Vector;

import org.apache.log4j.Logger;

import bpiwowar.argparser.checkers.OrChecker;
import bpiwowar.argparser.handlers.Handler;
import bpiwowar.argparser.handlers.HelpHandler;
import bpiwowar.argparser.holders.ObjectHolder;
import bpiwowar.argparser.utils.Pair;

/**
 * ArgParser is used to parse the command line arguments for a java application
 * program.
 * 
 * @author B. Piwowarski <benjamin@bpiwowar.net>
 */
public class ArgParser {
	// Our logger
	final static Logger logger = Logger.getLogger(ArgParser.class);

	/**
	 * Our options
	 */
	private ArrayList<Option> options = new ArrayList<Option>();

	/**
	 * Command line synopsis
	 */
	String synopsisString;

	/**
	 * Are the help options enabled?
	 */
	private boolean helpOptionsEnabled = true;

	/**
	 * Default stream for outputing messages
	 */
	private PrintStream printStream = System.err;

	/**
	 * Number of spaces for help
	 */
	int helpIndent = 24;

	String errMsg = null;

	/**
	 * List of methods to be called when command line has been parsed
	 */
	private ArrayList<Pair<Object, Method>> checkers = new ArrayList<Pair<Object, Method>>();

	/**
	 * List of argument to be parsed at the end of the command line
	 */
	private ArrayList<UnnamedOption> nonOptionRecords = new ArrayList<UnnamedOption>();

	/**
	 * Maps the groups to a set of options
	 */
	private Map<String, ArrayList<Option>> groups = new TreeMap<String, ArrayList<Option>>();

	/**
	 * If a handler implements this interface, it can be used to know if the
	 * handler can process a given field
	 * 
	 * @author Benjamin Piwowarski <benjamin@bpiwowar.net>
	 */
	static public interface TypeCheckerInterface {
		boolean isCompatible(Type type);
	}

	/**
	 * Creates an <code>ArgParser</code> with a synopsis string.
	 * 
	 * @param synopsisString
	 *            string that briefly describes program usage, for use by
	 *            {@link #getHelpMessage getHelpMessage}.
	 * @param objects
	 *            An optional list of objects to parse. It is possible to delay
	 *            parsing and use {@linkplain #addOptions(Object)} to add one by
	 *            one objects
	 */
	public ArgParser(String synopsisString, Object... objects) {
		this(synopsisString, true);
		// Add all the objects in the list
		for (Object object : objects)
			addOptions(object);
	}

	/**
	 * Creates an <code>ArgParser</code> with a synopsis string. The help
	 * options <code>-help</code> and <code>-?</code> are added if
	 * <code>defaultHelp</code> is true.
	 * 
	 * @param synopsisString
	 *            string that briefly describes program usage, for use by
	 *            {@link #getHelpMessage getHelpMessage}.
	 * @param defaultHelp
	 *            if true, adds the default help options
	 * @see ArgParser#getSynopsisString
	 * @see ArgParser#getHelpMessage
	 */
	public ArgParser(String synopsisString, boolean defaultHelp) {
		this.synopsisString = synopsisString;
		if (defaultHelp) {
			Option option = new Option();
			option.optionNames.add(new OptionName("--help"));
			option.optionNames.add(new OptionName("-?"));
			option.helpMessage = "displays help information";
			option.handler = new HelpHandler();
			addOption(option);
		}
	}

	/**
	 * Returns the synopsis string used by the parser. The synopsis string is a
	 * short description of how to invoke the program, and usually looks
	 * something like
	 * <p>
	 * <prec> &quot;java somepackage.SomeClass [options] files ...&quot;
	 * &lt;/prec&gt;
	 * 
	 * <p>
	 * It is used in help and error messages.
	 * 
	 * @return synopsis string
	 * @see ArgParser#setSynopsisString
	 * @see ArgParser#getHelpMessage
	 * 
	 */
	public String getSynopsisString() {
		return synopsisString;
	}

	/**
	 * Sets the synopsis string used by the parser.
	 * 
	 * @param s
	 *            The new synopsis string
	 * @see ArgParser#getSynopsisString
	 * @see ArgParser#getHelpMessage
	 */
	public void setSynopsisString(String s) {
		synopsisString = s;
	}

	/**
	 * Indicates whether or not help options are enabled.
	 * 
	 * @return true if help options are enabled
	 * @see ArgParser#setHelpOptionsEnabled
	 * @see ArgParser#addOption
	 */
	public boolean getHelpOptionsEnabled() {
		return helpOptionsEnabled;
	}

	/**
	 * Enables or disables help options. Help options are those associated with
	 * a conversion code of <code>%h</code>. If help options are enabled, and a
	 * help option is matched, then the string produced by
	 * {@link #getHelpMessage getHelpMessage} is printed to the default print
	 * stream and the program exits with code 0. Otherwise, arguments which
	 * match help options are ignored.
	 * 
	 * @param enable
	 *            enables help options if <code>true</code>.
	 * @see ArgParser#getHelpOptionsEnabled
	 * @see ArgParser#addOption
	 * @see ArgParser#setDefaultPrintStream
	 */
	public void setHelpOptionsEnabled(boolean enable) {
		helpOptionsEnabled = enable;
	}

	/**
	 * Returns the default print stream used for outputting help and error
	 * information.
	 * 
	 * @return default print stream
	 * @see ArgParser#setDefaultPrintStream
	 */
	public PrintStream getDefaultPrintStream() {
		return printStream;
	}

	/**
	 * Sets the default print stream used for outputting help and error
	 * information.
	 * 
	 * @param stream
	 *            new default print stream
	 * @see ArgParser#getDefaultPrintStream
	 */
	public void setDefaultPrintStream(PrintStream stream) {
		printStream = stream;
	}

	/**
	 * Gets the indentation used by {@link #getHelpMessage getHelpMessage}.
	 * 
	 * @return number of indentation columns
	 * @see ArgParser#setHelpIndentation
	 * @see ArgParser#getHelpMessage
	 */
	public int getHelpIndentation() {
		return helpIndent;
	}

	/**
	 * Sets the indentation used by {@link #getHelpMessage getHelpMessage}. This
	 * is the number of columns that an option's help information is indented.
	 * If the option's name and value information can fit within this number of
	 * columns, then all information about the option is placed on one line.
	 * Otherwise, the indented help information is placed on a separate line.
	 * 
	 * @param indent
	 *            number of indentation columns
	 * @see ArgParser#getHelpIndentation
	 * @see ArgParser#getHelpMessage
	 */
	public void setHelpIndentation(int indent) {
		helpIndent = indent;
	}

	/**
	 * Automatically add options for an object
	 * 
	 * @param object
	 *            The object instance to analyse
	 * @param recursive
	 *            If true, explores the super classes of this object, and the
	 *            members annotated {@linkplain ArgumentClass}
	 * @author B. Piwowarski
	 * @throws InvalidHolderException
	 *             If an error occurs while reading the object annotations
	 */
	public void addOptions(Object object, boolean recursive)
			throws InvalidHolderException {
		addOptions(object, recursive, true, "", "");
	}

	/**
	 * Add options for an object
	 * 
	 * @param object
	 *            The object to analyse
	 * @param recursive
	 *            Should we analyse the ancestors of this object
	 * @param activated
	 * @param prefix
	 * @param group
	 * @throws InvalidHolderException
	 */
	public void addOptions(Object object, boolean recursive, boolean activated,
			String prefix, String group) throws InvalidHolderException {

		Class<?> objectClass = object.getClass();

		while (objectClass != null) {
			// add the options
			addOptions(object, objectClass, activated, prefix, group);

			// If not recursive, stop now
			if (!recursive)
				break;

			objectClass = objectClass.getSuperclass();
		}

	}

	/**
	 * Add options from an annotated object instance, disregarding the
	 * superclasses
	 * 
	 * @param object
	 *            The object to explore
	 * @throws InvalidHolderException
	 *             if there is an error while parsing annotations
	 */
	public ArgParser addOptions(Object object) throws InvalidHolderException {
		addOptions(object, object.getClass());
		return this;
	}

	/**
	 * Add an option
	 * 
	 * @param option
	 *            The option to add
	 */
	public void addOption(Option option) {
		options.add(option);
	}

	/**
	 * Main method for adding options - this is where all is processed
	 * 
	 * @param object
	 *            The object to explore
	 * @param objectClass
	 *            The class of the object to explore
	 * @param defaultGroup
	 *            The default group
	 * @param defaultPrefix
	 *            The default prefix
	 * @param defaultActivation
	 *            The default activation
	 * @throws InvalidHolderException
	 *             if there is an error while parsing annotations
	 */
	private void addOptions(Object object, Class<?> objectClass)
			throws InvalidHolderException {
		addOptions(object, objectClass, true, "", "");
	}

	public void addOptions(Object object, Class<?> objectClass,
			boolean defaultActivated, String prefix, String defaultGroup)
			throws InvalidHolderException {

		// --- Get the list of methods with the @ArgumentPostProcessor
		// annotation
		for (Method method : objectClass.getDeclaredMethods()) {
			if (method.getAnnotation(ArgumentPostProcessor.class) != null) {
				checkers.add(new Pair<Object, Method>(object, method));
			}
		}

		// --- Explores the fields of the object
		List<AccessibleObject> members = new ArrayList<AccessibleObject>();
		members.addAll(ListAdaptator.create(objectClass.getDeclaredFields()));
		members.addAll(ListAdaptator.create(objectClass.getDeclaredMethods()));

		for (AccessibleObject member : members) {

			// --- Ordered argument case
			OrderedArgument orderedArgument = member
					.getAnnotation(OrderedArgument.class);

			if (orderedArgument != null) {
				Handler handler = Handlers.getHandler(object, member,
						orderedArgument.handler(), null);
				nonOptionRecords.add(new UnnamedOption(orderedArgument.name(),
						handler, orderedArgument.required()));
				continue;
			}

			Argument argument = member.getAnnotation(Argument.class);

			if (argument == null) {
				// --- check if we should follow up
				ArgumentClass followUp = member
						.getAnnotation(ArgumentClass.class);
				if (followUp != null && member instanceof Field) {
					Field field = (Field) member;
					try {
						boolean accessible = true;
						if (!member.isAccessible()) {
							member.setAccessible(true);
							accessible = false;
						}
						Object followUpObject = field.get(object);
						if (followUpObject == null) {
							// Create the object in case we have a null value
							Constructor<?> constructor = field.getType()
									.getConstructor(new Class<?>[] {});
							boolean constructAccess = constructor
									.isAccessible();
							if (!constructAccess)
								constructor.setAccessible(true);
							followUpObject = field.getType().newInstance();
							field.set(object, followUpObject);
							if (!constructAccess)
								constructor.setAccessible(false);
						}

						String followUpPrefix = String.format(prefix.isEmpty()
								|| followUp.prefix().isEmpty() ? "%s%s"
								: "%s-%s", prefix, followUp.prefix());

						addOptions(
								followUpObject,
								true,
								followUp.activated(),
								followUpPrefix,
								defaultGroup != null && !defaultGroup.isEmpty() ? defaultGroup
										: followUp.group());
						if (!accessible)
							member.setAccessible(false);

					} catch (InstantiationException e) {
						throw new InvalidHolderException(e);
					} catch (IllegalAccessException e) {
						throw new InvalidHolderException(e);
					} catch (SecurityException e) {
						throw new InvalidHolderException(e);
					} catch (NoSuchMethodException e) {
						throw new InvalidHolderException(e);
					}
				}
				continue;
			}

			// --- We have one annotation

			// Get the handler
			Handler handler = Handlers.getHandler(object, member,
					argument.handler(), argument.checkers());
			if (prefix != null)
				handler.setPrefix(prefix);

			try {

				// For the moment, if names is not

				if (argument.names().length + argument.name().length == 0)
					throw new InvalidHolderException("names() attribute of "
							+ member + " is empty");

				Option option = new Option();
				option.helpMessage = argument.help();
				option.handler = handler;
				option.setMandatory(argument.required().length == 0 ? false
						: argument.required()[0]);
				option.setActive(argument.activated().length == 0 ? defaultActivated
						: argument.activated()[0]);

				// Add the option to a group (null group is the default)
				String group = argument.group();
				if (group.equals(""))
					group = defaultGroup;
				ArrayList<Option> list = groups.get(group);
				if (list == null)
					groups.put(group, list = new ArrayList<Option>());
				list.add(option);

				// --- Add the intervals
				// FIXME implement this as handler specific annotations
				handler.processAnnotations();

				// newish implementation: should be taken out of this
				CheckNumericIntervals numericIntervals = member
						.getAnnotation(CheckNumericIntervals.class);
				if (numericIntervals != null)
					for (String interval : numericIntervals.value()) {
						OrChecker rangesChecker = new OrChecker();
						handler.updateRangeChecker(rangesChecker, "{"
								+ interval + "}");
						handler.add(rangesChecker);
					}

				// Old implementation
				for (String interval : argument.intervals()) {
					OrChecker rangesChecker = new OrChecker();
					handler.updateRangeChecker(rangesChecker, "{" + interval
							+ "}");
					handler.add(rangesChecker);
				}

				// --- Add option name(s)
				for (String optionName : argument.names())
					option.optionNames.add(new OptionName(optionName));

				for (String longName : argument.name())
					option.optionNames.add(new OptionName("--"
							+ (prefix == null || prefix.equals("") ? ""
									: prefix + "-") + longName));

				// If we have to activate a group when a value is set
				ArgumentGroupActivator activate = member
						.getAnnotation(ArgumentGroupActivator.class);
				if (activate != null) {
					option.setActivators(ListAdaptator.create(activate.value()));
				}

				addOption(option);

			} catch (IllegalArgumentException e) {
				IllegalArgumentException exception = new IllegalArgumentException(
						e.getMessage() + " while adding option "
								+ member.toString() + "", e.getCause());
				exception.setStackTrace(e.getStackTrace());
				throw exception;
			} catch (StringScanException e) {
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 * Concatenate a list of string with "-" separator
	 * 
	 * @param strings
	 *            An array of strings
	 * @return The concatenation (or null if empty)
	 */
	static public String concatenatePrefixes(String... strings) {
		StringBuilder sb = new StringBuilder();
		for (String prefix : strings) {
			if (prefix != null && prefix.length() > 0) {
				if (sb.length() > 0)
					sb.append("-");
				sb.append(prefix);
			}
		}

		if (sb.length() == 0)
			return null;

		return sb.toString();
	}

	/**
	 * Get the corresponding option for a given argument name
	 * 
	 * @param arg
	 *            The option in the command line
	 * @param ndescHolder
	 * @return
	 */
	private Option getOption(String arg, ObjectHolder<OptionName> ndescHolder) {
		Option matchingOption = null;
		for (Option option : options) {
			for (OptionName description : option.optionNames) {
				if (arg.equals(description.name) && option.isActivated()) {
					if (matchingOption != null)
						throw new RuntimeException(String.format(
								"At least two options have the same name (%s)",
								description.name));

					if (ndescHolder != null) {
						ndescHolder.value = description;
					}
					matchingOption = option;
				}
			}
		}
		return matchingOption;
	}

	/**
	 * Transforms a string into a list of arguments
	 * @param vec
	 * @param s
	 * @param allowQuotedStrings
	 * @throws StringScanException
	 */
	static void stringToArgs(List<String> vec, String s,
			boolean allowQuotedStrings) throws StringScanException {
		StringScanner scanner = new StringScanner(s);
		scanner.skipWhiteSpace();
		while (!scanner.atEnd()) {
			if (allowQuotedStrings) {
				vec.add(scanner.scanString());
			} else {
				vec.add(scanner.scanNonWhiteSpaceString());
			}
			scanner.skipWhiteSpace();
		}
	}

	/**
	 * Reads in a set of strings from a reader and prepends them to an argument
	 * list. Strings are delimited by either whitespace or double quotes
	 * <code>"</code>. The character <code>#</code> acts as a comment character,
	 * causing input to the end of the current line to be ignored.
	 * 
	 * @param reader
	 *            Reader from which to read the strings
	 * @param args
	 *            Initial set of argument values. Can be specified as
	 *            <code>null</code>.
	 * @throws IOException
	 *             if an error occured while reading.
	 */
	public static String[] prependArgs(Reader reader, String[] args)
			throws IOException {
		if (args == null) {
			args = new String[0];
		}
		LineNumberReader lineReader = new LineNumberReader(reader);
		Vector<String> vec = new Vector<String>(100, 100);
		String line;
		int i, k;

		while ((line = lineReader.readLine()) != null) {
			int commentIdx = line.indexOf("#");
			if (commentIdx != -1) {
				line = line.substring(0, commentIdx);
			}
			try {
				stringToArgs(vec, line, /* allowQuotedStings= */true);
			} catch (StringScanException e) {
				throw new IOException("malformed string, line "
						+ lineReader.getLineNumber());
			}
		}
		String[] result = new String[vec.size() + args.length];
		for (i = 0; i < vec.size(); i++) {
			result[i] = vec.get(i);
		}
		for (k = 0; k < args.length; k++) {
			result[i++] = args[k];
		}
		return result;
	}

	/**
	 * Reads in a set of strings from a file and prepends them to an argument
	 * list. Strings are delimited by either whitespace or double quotes
	 * <code>"</code>. The character <code>#</code> acts as a comment character,
	 * causing input to the end of the current line to be ignored.
	 * 
	 * @param file
	 *            File to be read
	 * @param args
	 *            Initial set of argument values. Can be specified as
	 *            <code>null</code>.
	 * @throws IOException
	 *             if an error occured while reading the file.
	 */
	public static String[] prependArgs(File file, String[] args)
			throws IOException {
		if (args == null) {
			args = new String[0];
		}
		if (!file.canRead()) {
			return args;
		}
		try {
			return prependArgs(new FileReader(file), args);
		} catch (IOException e) {
			throw new IOException("File " + file.getName() + ": "
					+ e.getMessage());
		}
	}

	/**
	 * Sets the parser's error message.
	 * 
	 * @param s
	 *            Error message
	 */
	protected void setError(String msg) {
		errMsg = msg;
	}

	/**
	 * Prints an error message, along with a pointer to help options, if
	 * available, and causes the program to exit with code 1.
	 */
	public void printErrorAndExit(String format, Object... args) {
		if (helpOptionsEnabled)
			format += "%nUse -? or --help for help information";
		if (printStream != null) {
			printStream.format(format, args);
			printStream.println();
		}
		System.exit(1);
	}

	/**
	 * Matches arguments within an argument list.
	 * 
	 * <p>
	 * In the event of an erroneous or unmatched argument, the method prints a
	 * message and exits the program with code 1.
	 * 
	 * <p>
	 * If help options are enabled and one of the arguments matches a help
	 * option, then the result of {@link #getHelpMessage getHelpMessage} is
	 * printed to the default print stream and the program exits with code 0. If
	 * help options are not enabled, they are ignored.
	 * 
	 * @param args
	 *            argument list
	 * @throws ArgParserException
	 * @see ArgParser#getDefaultPrintStream
	 */
	public void matchAllArgs(String... args) throws ArgParserException {
		matchAllArgs(args, 0, ArgParserOption.EXIT_ON_UNMATCHED,
				ArgParserOption.EXIT_ON_ERROR);
	}

	/**
	 * When we finished parsing the command line, proceeds to a check of the
	 * arguments by colling the checkers
	 * 
	 * @throws ArgParseException
	 *             If something went wrong
	 */
	private void check() throws ArgParseException {
		for (Pair<Object, Method> pair : checkers) {
			try {
				boolean isAccessible = pair.getSecond().isAccessible();
				if (!isAccessible)
					pair.getSecond().setAccessible(true);
				pair.getSecond().invoke(pair.getFirst(), new Object[] {});
				if (!isAccessible)
					pair.getSecond().setAccessible(isAccessible);

			} catch (IllegalAccessException e) {
				logger.warn("Field "
						+ pair.getSecond()
						+ " cannot be accessed (hint: add the \"public\" keyword");
				throw new RuntimeException(e);
			} catch (Exception e) {
				if (e.getCause() instanceof ArgParseException)
					throw ((ArgParseException) e.getCause());
				throw new RuntimeException(e);
			}
		}
	}

	/**
	 * Matches a command line
	 * 
	 * @param args
	 * @param parseFlags
	 * @return
	 * @throws ArgParserException
	 */
	public String[] matchAllArgs(String[] args,
			EnumSet<ArgParserOption> parseFlags) throws ArgParserException {
		return matchAllArgs(args, 0, parseFlags);
	}

	public String[] matchAllArgs(String[] args, int idx,
			ArgParserOption... exitFlags) throws ArgParserException {
		EnumSet<ArgParserOption> parseFlags = EnumSet
				.noneOf(ArgParserOption.class);
		for (ArgParserOption option : exitFlags)
			parseFlags.add(option);
		return matchAllArgs(args, idx, parseFlags);
	}

	/**
	 * Matches arguments within an argument list and returns those which were
	 * not matched. The matching starts at a location in <code>args</code>
	 * specified by <code>idx</code>, and unmatched arguments are returned in a
	 * String array.
	 * 
	 * <p>
	 * In the event of an erroneous argument, the method either prints a message
	 * and exits the program (if {@link #EXIT_ON_ERROR} is set in
	 * <code>exitFlags</code>) or terminates the matching and creates a error
	 * message that can be retrieved by {@link #getErrorMessage}.
	 * 
	 * <p>
	 * In the event of an umatched argument, the method will print a message and
	 * exit if {@link #EXIT_ON_UNMATCHED} is set in <code>errorFlags</code>.
	 * Otherwise, the unmatched argument will be appended to the returned array
	 * of unmatched values, and the matching will continue at the next location.
	 * 
	 * <p>
	 * If help options are enabled and one of the arguments matches a help
	 * option, then the result of {@link #getHelpMessage getHelpMessage} is
	 * printed to the the default print stream and the program exits with code
	 * 0. If help options are not enabled, then they will not be matched.
	 * 
	 * @param args
	 *            argument list
	 * @param idx
	 *            starting location in list
	 * @param exitFlags
	 *            conditions causing the program to exit. Should be an or-ed
	 *            combination of {@link #ParseFlag.EXIT_ON_ERROR}, {@link #ParseFlag.STOP_FIRST_UNMATCHED} or
	 *            {@link #ParseFlag.EXIT_ON_UNMATCHED}.
	 * @return array of arguments that were not matched, or <code>null</code> if
	 *         all arguments were successfully matched
	 * @throws ArgParserException
	 * @throws StringScanException
	 * @see ArgParserOption
	 * @see ArgParser#getErrorMessage
	 * @see ArgParser#getDefaultPrintStream
	 */
	public String[] matchAllArgs(String[] args, int idx,
			EnumSet<ArgParserOption> exitFlags) throws ArgParserException {
		ArrayList<String> unmatched = new ArrayList<String>(10);

		int noIdx = 0;

		while (idx < args.length) {
			debug(logger, "Processing argument %d/%d", idx + 1, args.length);
			try {
				// Read the command
				final int oldIdx = idx;
				idx = matchArg(args, idx);
				if (oldIdx == idx) {
					logger.debug(args[idx] + " was not matched");
					if (noIdx < nonOptionRecords.size()) {
						UnnamedOption option = nonOptionRecords.get(noIdx);
						idx = option.handler.parse(this, args, idx);
						noIdx++;
						continue;
					} else if (exitFlags
							.contains(ArgParserOption.EXIT_ON_UNMATCHED)) {
						throw new ArgParserException("Unrecognized argument: "
								+ args[idx]);
					} else if (exitFlags
							.contains(ArgParserOption.EXCEPTION_ON_UNMATCHED)) {
						throw new UnmatchedArgumentException(
								"Unrecognized argument: " + args[idx]);
					} else {
						unmatched.add(args[idx]);
						idx++;
						if (exitFlags
								.contains(ArgParserOption.STOP_FIRST_UNMATCHED)) {
							while (idx < args.length) {
								unmatched.add(args[idx]);
								idx++;
							}
							check();
							return unmatched.toArray(new String[0]);
						}

					}
				}

			} catch (UnmatchedArgumentException e) {
				throw e;
			} catch (IllegalArgumentValue e) {
				if (exitFlags.contains(ArgParserOption.EXIT_ON_ERROR)) {
					printErrorAndExit(e.getMessage());
				} else if (exitFlags
						.contains(ArgParserOption.EXCEPTION_ON_ERROR)) {
					throw e;
				}
				break;
			}
		}

		args = unmatched.toArray(new String[0]);

		if (noIdx < nonOptionRecords.size()) {
			for (int i = noIdx; i < nonOptionRecords.size(); i++)
				if (nonOptionRecords.get(i).required)
					throw new RequiredArgumentMissing(
							"Argument '%s' was not matched in the command line",
							nonOptionRecords.get(i).name);
		}

		// Check the arguments
		check();

		// Check for mandatory arguments
		for (Option record : options) {
			if (record.isMandatory() && record.isActivated()
					&& !record.isProcessed()) 
				throw new RequiredArgumentMissing("Option " + record.getNames()
						+ " was not given and is mandatory");
		}

		return args;
	}

	public static void debug(Logger logger, String format, Object... args) {
		if (logger.isDebugEnabled())
			logger.debug(String.format(format, args));
	}

	/**
	 * Matches one option starting at a specified location in an argument list.
	 * The method returns the location in the list where the next match should
	 * begin.
	 * 
	 * <p>
	 * In the event of an erroneous argument, the method throws an
	 * {@link bpiwowar.argparser.ArgParseException ArgParseException} with an
	 * appropriate error message. This error message can also be retrieved using
	 * {@link #getErrorMessage getErrorMessage}.
	 * 
	 * <p>
	 * In the event of an umatched argument, the method will return idx
	 * </p>
	 * 
	 * <p>
	 * If help options are enabled and the argument matches a help option, then
	 * the result of {@link #getHelpMessage getHelpMessage} is printed to the
	 * the default print stream and the program exits with code 0. If help
	 * options are not enabled, then they are ignored.
	 * </p>
	 * 
	 * @param args
	 *            argument list
	 * @param idx
	 *            location in list where match should start
	 * @return location in list where next match should start
	 * @throws ArgParserException
	 * @throws StringScanException
	 * @see ArgParser#setDefaultPrintStream
	 * @see ArgParser#getHelpOptionsEnabled
	 * @see ArgParser#getErrorMessage
	 * @see ArgParser#getUnmatchedArgument
	 */
	public int matchArg(String[] args, int idx) throws ArgParserException {
		setError(null);
		try {

			// Retrieve the handler for args[idx]
			ObjectHolder<OptionName> ndescHolder = new ObjectHolder<OptionName>();
			final String argName = args[idx];
			Option option = getOption(argName, ndescHolder);

			// no match
			if (option == null) {
				return idx;
			}

			logger.debug("Handler is " + option.handler + " for " + argName);

			// Help wanted
			if (option.handler.isHelpOption()) {
				if (helpOptionsEnabled) {
					printStream.println(getHelpMessage());
					System.exit(0);
				} else {
					return idx + 1;
				}
			}

			try {
				// Parse and check
				idx = option.handler.parse(this, args, idx + 1);

				// Activate some groups if needed
				for (String group : option.toActivate)
					activate(group);
			} catch (RuntimeException e) {
				logger.error("Error while processing with handler "
						+ option.handler + " (argument " + argName + ")");
				throw e;
			} catch (StringScanException e) {
				logger.error("Error while processing with handler "
						+ option.handler + " (argument " + argName + ")");
				throw new ArgParseException(e,
						"String scan exception [deprecated]");
			}
			logger.debug(String.format("Option %s has been processed",
					option.getNames()));
			option.setProcessed(true);
		} catch (ArgParseException e) {
			setError(e.getMessage());
			throw e;
		}
		return idx;
	}

	private String spaceString(int n) {
		StringBuffer sbuf = new StringBuffer(n);
		for (int i = 0; i < n; i++) {
			sbuf.append(' ');
		}
		return sbuf.toString();
	}

	/**
	 * Returns a string describing the allowed options in detail.
	 * 
	 * @return help information string.
	 */
	public String getHelpMessage() {
		StringBuilder s = new StringBuilder();

		s.append("Usage: " + synopsisString);
		for (UnnamedOption record : nonOptionRecords) {
			s.append(" <" + record.name + ">");
		}
		s.append('\n');
		s.append("Options include:\n\n");
		for (Entry<String, ArrayList<Option>> entry : groups.entrySet()) {

			String groupName = entry.getKey().equals("") ? "Default" : entry
					.getKey();
			s.append("\n[");
			s.append(groupName);
			s.append("]\n\n");

			for (Option option : entry.getValue()) {
				String optionInfo = option.getNames();

				s.append(optionInfo);
				if (option.helpMessage.length() > 0) {
					int pad = helpIndent - optionInfo.length();
					if (pad < 2) {
						s.append('\n');
						pad = helpIndent;
					}
					s.append(spaceString(pad));
					s.append(option.helpMessage);
				}
				s.append('\n');
			}
		}

		return s.toString();
	}

	/**
	 * Returns the parser's error message. This is automatically set whenever an
	 * error is encountered in <code>matchArg</code> or
	 * <code>matchAllArgs</code>, and is automatically set to <code>null</code>
	 * at the beginning of these methods.
	 * 
	 * @return error message
	 */
	public String getErrorMessage() {
		return errMsg;
	}

	public void activate(String groupId) {
		logger.debug("Activating group " + groupId);
		ArrayList<Option> options = groups.get(groupId);
		if (options != null)
			for (Option option : options)
				option.setActive(true);
	}

}
