/**
 * 
 */
package bpiwowar.argparser;

import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;

/**
 * Used to mark a variable as a set of other options. If a class <code>A</code>
 * contains the following code:
 * <p>
 * 
 * <pre>
 * <code>
 * class A {
 * ...
 * <span>@ArgumentClass</span>
 * OptionsClass aClass;
 * ...
 * }
 * </code>
 * </pre>
 * 
 * </p>
 * 
 * <p>
 * Calling {@linkplain ArgParser#addOptions(Object)} with argument an instance
 * of class A, will add the options of object <code>OptionsClass</code>.
 * </p>
 * 
 * <p>
 * The object will be created with the default constructor if not inialised
 * </p>
 * 
 * @author Benjamin Piwowarski <benjamin@bpiwowar.net>
 */
@Retention(value = RetentionPolicy.RUNTIME)
public @interface ArgumentClass {
	/**
	 * The default group for the options
	 * 
	 * @return
	 */
	String group() default "";

	/**
	 * The defaut activated value
	 * 
	 * @return
	 */
	boolean activated() default true;

	/**
	 * Default prefix to add to the options
	 * 
	 * @return
	 */
	String prefix() default "";

	/**
	 * Is the arguments in this class required?
	 * 
	 * The semantics is as follows. If required() is
	 * <dl>
	 * <dt>true</dt>
	 * <dd>
	 * in the parsed class, any required argument will be required, the others
	 * won't.</dd>
	 * 
	 * <dt>false</dt>
	 * <dd>
	 * 
	 * In case this is false, then if on the command line one argument belonging
	 * to this class appears, all the required arguments become required.</dd>
	 * </dl>
	 */
	boolean required() default false;

	/**
	 * The help for this set of options
	 */
	String help() default "";
}
