package bpiwowar.argparser;

import java.lang.reflect.Array;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * Helper methods for generics
 * 
 * @author bpiwowar
 */
public class GenericHelper {
	final public static <T> ArrayList<T> newArrayList() {
		return new ArrayList<T>();
	}

	final public static <K, V> TreeMap<K, V> newTreeMap() {
		return new TreeMap<K, V>();
	}

	final public static <K, V> HashMap<K, V> newHashMap() {
		return new HashMap<K, V>();
	}

	final public static <T> HashSet<T> newHashSet() {
		return new HashSet<T>();
	}

	/**
	 * Check if the generic types of an implemented superclass/interface match
	 * 
	 * <p>
	 * 
	 * <pre>
	 * GenericHelper.checkGenerics(Collection.class, gtype, String.class);
	 * </pre>
	 * 
	 * </p>
	 * 
	 * @param candidate
	 * @param gtype
	 * @param classes
	 * @return
	 */
	static public boolean checkGenerics(Class<?> candidate, Type gtype,
			@SuppressWarnings("rawtypes") Class... classes) {
		if (gtype instanceof ParameterizedType) {
			// list the raw type information
			ParameterizedType ptype = (ParameterizedType) gtype;
			Type rtype = ptype.getRawType();
			if (rtype instanceof Class<?>) {
				Class<?> ctype = (Class<?>) rtype;
				if (Collection.class.isAssignableFrom(ctype)) {
					// list the actual type arguments
					Type[] targs = ptype.getActualTypeArguments();
					if (targs.length != classes.length)
						return false;
					for (int i = 0; i < targs.length; i++) {
						if (!targs[i].equals(classes[i]))
							return false;
					}
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Reflection helper - give the generic parameters of a class.
	 * 
	 * @param candidate
	 *            The class implementing the superclass or superinterface
	 * @param implemented
	 * @return a type array of the type arguments of a superclass or a
	 *         superinterface
	 */
	public static Type[] getActualTypeArguments(Class<?> candidate,
			Class<?> implemented) {

		// We are not a derived class ?
		if (!implemented.isAssignableFrom(candidate)) {
			return null;
		}

		// Get the list of candidates
		final boolean isInterface = implemented.isInterface();
		Class<?> current = candidate;

		do {
			final Type[] list = isInterface ? current.getGenericInterfaces()
					: new Type[] { current.getGenericSuperclass() };
			for (Type anInterface : list) {
				if (anInterface instanceof ParameterizedType) {
					ParameterizedType ptype = (ParameterizedType) anInterface;
					if (ptype.getRawType() == implemented) {
						return ptype.getActualTypeArguments();
					}
				}
			}

			current = current.getSuperclass();
		} while (!isInterface && current != null);

		// Should not happen!
		throw new RuntimeException(
				String.format(
						"Should not happen: we did not find the generic type of %s in %s",
						implemented, current));
	}

	public static <T> TreeSet<T> newTreeSet() {
		return new TreeSet<T>();
	}

	public static <T> ArrayList<T> newArrayList(int length) {
		return new ArrayList<T>(length);
	}

	@SuppressWarnings("unchecked")
	final public static <T, U extends T> U[] newArray(Class<T> theClass,
			int length) {
		return (U[]) Array.newInstance(theClass, length);
	}

}
