package bpiwowar.argparser;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;

import org.apache.log4j.Level;

import bpiwowar.argparser.ArgParser.TypeCheckerInterface;
import bpiwowar.argparser.checkers.ValueChecker;
import bpiwowar.argparser.handlers.EnumHandler;
import bpiwowar.argparser.handlers.Handler;
import bpiwowar.argparser.handlers.HandlerInformation;
import bpiwowar.argparser.handlers.MethodCallHandler;
import bpiwowar.argparser.holders.ObjectHolder;
import bpiwowar.argparser.utils.AnnotatedElementAnnotationMap;
import bpiwowar.argparser.utils.AnnotationMap;
import bpiwowar.argparser.utils.Introspection;
import bpiwowar.argparser.utils.Pair;

/**
 * This class is is a utility class for searching for handlers
 * 
 * @author B. Piwowarski
 */
public class Handlers {
	final static private Logger logger = Logger.getLogger();

	private static void log(Level level, String format, Object... objects) {
		logger.log(level, String.format(format, objects));
	}

	/**
	 * The list of potential handlers and the class that checks for their
	 * adequacy
	 */
	static private ArrayList<Pair<Class<? extends Handler>, TypeCheckerInterface>> defaultHandlers = new ArrayList<Pair<Class<? extends Handler>, TypeCheckerInterface>>();

	// --- Initialize the handlers ---
	static {
		ArrayList<Class<?>> classes = Introspection.getClasses(
				new Introspection.Checker() {
					public boolean accepts(Class<?> aClass) {
						return true;
					}
				}, ArgParser.class.getPackage().getName() + ".handlers", -1);
		for (Class<?> c : classes) {
			logger.debug("Holders list: examining " + c + " "
					+ Handler.class.isAssignableFrom(c));
			if (Handler.class.isAssignableFrom(c))
				try {
					HandlerInformation wrapperType = c
							.getAnnotation(HandlerInformation.class);
					TypeCheckerInterface checker = null;

					if (wrapperType != null) {
						if (wrapperType.value().length == 1
								&& TypeCheckerInterface.class
										.isAssignableFrom(wrapperType.value()[0]))
							checker = (TypeCheckerInterface) wrapperType
									.value()[0].newInstance();
						else
							checker = new TypeChecker(wrapperType.value());
					}

					if (checker != null) {
						@SuppressWarnings("unchecked")
						Class<? extends Handler> c2 = (Class<? extends Handler>) c;
						defaultHandlers
								.add(new Pair<Class<? extends Handler>, TypeCheckerInterface>(
										c2, checker));
						logger.debug("Adding wrapper " + c + " with checker "
								+ checker);
					}
				} catch (Exception e) {
					logger.debug("Cannot check object " + c.getName() + ": "
							+ e);
				}
		}
	}

	/**
	 * Get a handler for a given object member
	 * 
	 * @param object
	 * @param member
	 * @param annotationHandler
	 * @return A handler
	 * @throws InvalidHolderException
	 *             If the assigned handler cannot handle the object type
	 */
	static Handler getHandler(Object object, AccessibleObject member,
			Class<? extends Handler> annotationHandler,
			Class<? extends ValueChecker>[] checkers)
			throws InvalidHolderException {

		// The member is a field
		if (member instanceof Field) {
			final Field field = (Field) member;
			return getHandler(object, (Field) member, field.getGenericType(),
					new AnnotatedElementAnnotationMap(field.getType()),
					annotationHandler, checkers);
		}

		// The member is a method
		else if (member instanceof Method)
			return getHandler(object, (Method) member, annotationHandler);

		throw new RuntimeException("Cannnot process member of type "
				+ member.getClass());
	}

	/**
	 * Get the handler for a method
	 * 
	 * @param object
	 * @param method
	 * @param annotationHandler
	 * @return
	 * @throws InvalidHolderException
	 */
	private static Handler getHandler(Object object, Method method,
			Class<? extends Handler> annotationHandler)
			throws InvalidHolderException {
		try {
			// Construct the handler
			Type[] parameters = method.getGenericParameterTypes();
			Annotation[][] pAnnotations = method.getParameterAnnotations();
			final int length = parameters.length;
			final Handler[] handlers = new Handler[length];
			final ObjectHolder<?>[] holders = new ObjectHolder[length];
			log(Level.DEBUG, "Processing method %s (%d arguments)", method,
					length);

			// Parse every method argument
			for (int i = 0; i < length; i++) {
				final ListAnnotationMap a = new ListAnnotationMap(
						pAnnotations[i]);
				final Argument pa = a.getAnnotation(Argument.class);
				holders[i] = new ObjectHolder<Object>();

				log(Level.DEBUG,
						"Searching a handler for the %dth argument with annotation %s, and type %s",
						i + 1, pa, parameters[i]);

				// Set the handler
				handlers[i] = getHandler(holders[i],
						holders[i].getValueField(), parameters[i], a,
						pa != null ? pa.handler() : Handler.class,
						pa != null ? pa.checkers() : null);

			}

			// --- Setting the overall handler
			ArgumentRegexp regexp = method.getAnnotation(ArgumentRegexp.class);

			return new MethodCallHandler(object, method, handlers, holders,
					regexp != null ? regexp.value() : null);
		} catch (RuntimeException e) {
			throw e;
		} catch (Throwable e) {
			throw new InvalidHolderException(e);
		}

	}

	/**
	 * Get the class of the type
	 * 
	 * @param type
	 *            A given type
	 * @return The class
	 */
	static Class<?> getTypeClass(Type type) {
		if (type instanceof Class<?>)
			return (Class<?>) type;
		if (type instanceof ParameterizedType)
			return (Class<?>) ((ParameterizedType) type).getRawType();
		throw new RuntimeException();
	}

	/**
	 * Get the handler for an object member (either a method or a field)
	 * 
	 * @param object
	 *            The object containing the field
	 * @param field
	 *            The field
	 * @param fieldType
	 *            The field type
	 * @param annotationHandler
	 *            The handler declared for the field (if any)
	 * @param checkers
	 *            The checkers for this handler
	 * @return The handler object If a handler object has been given
	 * @throws InvalidHolderException
	 *             If we did not find any holder
	 */
	static Handler getHandler(Object object, Field field, Type fieldType,
			AnnotationMap annotations,
			Class<? extends Handler> annotationHandler,
			Class<? extends ValueChecker>[] checkers)
			throws InvalidHolderException {

		log(Level.DEBUG, "Finding a handler for %s (type %s)", field, fieldType);
		Handler handler = null;

		// --- Try to find a handler class from field class annotations

		// Get the type of field
		Class<?> fieldClass = getTypeClass(fieldType);
		Class<? extends Handler> handlerClass = null;

		if (!annotationHandler.equals(Handler.class))
			handlerClass = annotationHandler;
		else if (annotations.getAnnotation(ArgumentHandler.class) != null)
			handlerClass = annotations.getAnnotation(ArgumentHandler.class)
					.value();

		// If we have a designated handler
		if (handlerClass != null) {
			try {
				Constructor<? extends Handler> constructor = handlerClass
						.getConstructor(new Class[] { Object.class, Field.class });
				handler = constructor
						.newInstance(new Object[] { object, field });
			} catch (SecurityException e) {
				throw new InvalidHolderException(
						"Security exception while searching for wrapper constructor of "
								+ annotationHandler, e);
			} catch (NoSuchMethodException e) {
				throw new InvalidHolderException(
						"No valid constructor for wrapper " + annotationHandler
								+ " for field " + field, e);
			} catch (Exception e) {
				throw new InvalidHolderException(
						"Error while constructing wrapper " + annotationHandler
								+ " for field " + field, e);
			}

		} else {
			if (Handler.class.isAssignableFrom(fieldClass)) {
				try {
					handler = (Handler) field.get(object);
				} catch (Exception e) {
					throw new InvalidHolderException(e);
				}
			}

			else {
				// --- Try to see if there is a method with @ArgumentProcessor
				// annotation
				for (Method method : fieldClass.getDeclaredMethods()) {
					final ArgumentProcessor argProcessorAnt = method
							.getAnnotation(ArgumentProcessor.class);
					if (argProcessorAnt != null) {
						try {

							// --- Setting the overall handler
							boolean accessible = field.isAccessible();
							if (!accessible)
								field.setAccessible(true);
							Object fieldObject = field.get(object);
							if (!accessible)
								field.setAccessible(false);

							handler = getHandler(fieldObject, method,
									annotationHandler);

						} catch (RuntimeException e) {
							throw e;
						} catch (Throwable e) {
							throw new InvalidHolderException(e);
						}

					}
				}

				// --- Otherwise, we try to find a holder
				if (handler != null) {
				} else if (Enum.class.isAssignableFrom(fieldClass)) {
					EnumHandler enumWrapper = new EnumHandler(object, field);
					handler = enumWrapper;
				} else {
					// Search through the handlers
					handler = getHandler(fieldType, object, field, fieldClass);
				}
			}
		}

		if (handler == null)
			throw new InvalidHolderException(String.format(
					"Cannot construct holder for type of: %s", fieldType));

		// --- Set the checkers
		if (checkers != null)
			for (Class<? extends ValueChecker> checker : checkers)
				try {
					log(Level.DEBUG, "Adding checker of class %s", checker);
					handler.add(checker.newInstance());
				} catch (IllegalAccessException e) {
					throw new RuntimeException(e);
				} catch (InstantiationException e) {
					throw new RuntimeException(e);
				}

		logger.debug("Found handler " + handler.getClass().getName() + " for "
				+ fieldType + " / " + field.getName());

		return handler;
	}

	/**
	 * Search for a registered handler for a given value
	 * 
	 * @param handler
	 * @param fieldName
	 * @param objectClass
	 *            The class for the holder
	 * @param fieldClass
	 *            Case of creation
	 * 
	 * @return A handler
	 * @throws InvalidHolderException
	 *             If an handler cannot be found
	 */
	private static Handler getHandler(Type type, Object object, Field field,
			Class<?> fieldClass) throws InvalidHolderException {
		logger.debug("Searching a handler for type " + type);
		for (Pair<Class<? extends Handler>, TypeCheckerInterface> w : defaultHandlers) {
			// Get the annotation
			logger.debug("Checking compatibility of " + w);
			if (w.getSecond().isCompatible(type)) {
				try {
					Constructor<? extends Handler> constructor = w.getFirst()
							.getConstructor(
									new Class<?>[] { Object.class, Field.class,
											Class.class });
					return constructor.newInstance(object, field, fieldClass);

				} catch (Exception e) {
				}

				try {
					Constructor<? extends Handler> constructor = w
							.getFirst()
							.getConstructor(
									new Class<?>[] { Object.class, Field.class });
					return constructor.newInstance(object, field);

				} catch (Exception e) {
					throw new InvalidHolderException(
							"Cannot create a new instance of handler %s", e,
							w.getFirst());
				}
			}
		}

		throw new InvalidHolderException(String.format(
				"Cannot find a handler for type %s (object %s, field %s)",
				type, object.getClass(), field));

	}

}
