package bpiwowar.argparser.handlers;

import static java.lang.String.format;

import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.TreeMap;

import org.apache.log4j.Logger;

import bpiwowar.argparser.ArgParseException;
import bpiwowar.argparser.ArgParser;
import bpiwowar.argparser.Argument;
import bpiwowar.argparser.ArgParser.TypeCheckerInterface;
import bpiwowar.argparser.utils.Introspection;
import bpiwowar.argparser.utils.Introspection.Checker;

/**
 * A class chooser handler
 * 
 * Example of use <code><![[CDATA[
 * 	public static enum StemmerEnum {
 * 
 * @ClassChooser.Choice(value = EnglishStemmer.class, name = "english") ENGLISH,
 * @ClassChooser.Choice(value = EnglishStemmer.class, name = "french") FRENCH,
 *                            };
 * 
 * @Argument(name = "stem", help =
 *                "If stem should be returned instead of the POS & LEMMA, give the stemmer here"
 *                , handler = ClassChooser.class)
 * @ClassChooser.Choices(StemmerEnum.class) AbstractSnowballTermProcessor
 *                                          stemmer; ]]> </code>
 * 
 * 
 * @author B. Piwowarski <benjamin@bpiwowar.net>
 */
@HandlerInformation(ClassChooser.TypeChecker.class)
public class ClassChooser extends GenericObjectsHandler {
	final static private Logger LOGGER = Logger.getLogger(ClassChooser.class);

	// Our type checker
	static public class TypeChecker implements TypeCheckerInterface {

		/**
		 * Returns true if the type if ? extends Collection<String>
		 * 
		 * @see bpiwowar.argparser.ArgParser.TypeCheckerInterface#isCompatible(java.lang.Class,
		 *      Field)
		 */
		public boolean isCompatible(Type gtype) {
			Class<?> aClass = null;
			if (gtype instanceof ParameterizedType) {
				// list the raw type information
				Type rtype = ((ParameterizedType) gtype).getRawType();
				;
				if (rtype instanceof Class<?>)
					aClass = (Class<?>) rtype;
			} else if (gtype instanceof Class<?>)
				aClass = (Class<?>) gtype;

			if (aClass != null) {
				return aClass.getAnnotation(Instances.class) != null;
			}
			return false;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return "Collection<String> checker";
		}

	}

	/**
	 * Defines a set of classes that should implement
	 * 
	 * @author B. Piwowarski <benjamin@bpiwowar.net>
	 */
	@Retention(value = RetentionPolicy.RUNTIME)
	public @interface Instances {
		/**
		 * A set of choices
		 */
		Class<?>[] value() default {};

		/**
		 * A set of packages to explore for instances
		 */
		String[] packages() default {};
	}

	@Retention(value = RetentionPolicy.RUNTIME)
	@Deprecated
	/**
	 * The prefix for sub-options
	 * 
	 * Use the prefix() 
	 */
	public @interface Prefix {
		String value();
	}

	/**
	 * Gives the reference to an enumeration of the possible choices. Use
	 * {@link #ClassChooser.Instances} to give a list of choices
	 * 
	 * @author B. Piwowarski <benjamin@bpiwowar.net>
	 */
	@Retention(value = RetentionPolicy.RUNTIME)
	static public @interface Choices {
		Class<? extends Enum<?>> value();
	};

	/**
	 * @author B. Piwowarski <benjamin@bpiwowar.net>
	 */
	@Retention(value = RetentionPolicy.RUNTIME)
	static public @interface Choice {
		/**
		 * The class (mandatory if the annotation is for an enum value)
		 */
		Class<?> value() default Object.class;

		/** A name to replace the enum value */
		String name() default "";

		/** Help for the value */
		String help() default "";

		/** Activate a group */
		String group() default "";

		/** Add a default prefix to the group */
		String groupPrefix() default "";
	}

	/**
	 * Information about a possible choice
	 * 
	 * @author Benjamin Piwowarski <benjamin@bpiwowar.net>
	 */
	static class Information {
		Class<?> theClass;
		String group;
		String groupPrefix;

		public Information(Class<?> theClass, String group, String groupPrefix) {
			this.theClass = theClass;
			this.group = group;
			this.groupPrefix = groupPrefix;
			LOGGER.debug(String.format("Adding option %s/%s/%s", theClass,
					group, groupPrefix));
		}

		public Information(Choice choice) {
			this(choice.value(), choice.group(), choice.groupPrefix());
		}
	}

	TreeMap<String, Information> map = new TreeMap<String, Information>();

	/**
	 * Defines the global prefix for any option defined within the instanciated
	 * class
	 */
	private String localPrefix;
	
	
	/**
	 * Defines the global prefix for any option defined within the instanciated
	 * class
	 */
	private String prefix;
	
	public void setPrefix(String prefix) {
		this.prefix = prefix;
	}
	
	
	/**
	 * Initialisation of the class chooser
	 * 
	 * @param object
	 * @param field
	 */
	public ClassChooser(Object object, Field field) {
		super(object, field);

		LOGGER
				.debug("ClassChooser for field " + field + " of object "
						+ object);

		// Get the potential prefix for further options
		Prefix prefixAnnotation = field.getAnnotation(Prefix.class);
		if (prefixAnnotation != null)
			localPrefix = prefixAnnotation.value();
		else {
			final Argument argumentAnnotation = field
					.getAnnotation(Argument.class);
			if (argumentAnnotation != null)
				localPrefix = argumentAnnotation.prefix();
		}

		LOGGER.debug("Global prefix is " + localPrefix);

		// Get the annotation
		Choices choicesClass = field.getAnnotation(ClassChooser.Choices.class);

		if (choicesClass == null) {
			// --- There is no choice class: we look into the object
			// class
			final Class<? extends Object> objectClass = field.getType();

			Instances instances = field.getAnnotation(Instances.class);
			if (instances == null)
				instances = objectClass.getAnnotation(Instances.class);

			if (instances != null) {
				// Using instances
				for (Class<?> aClass : instances.value())
					if (!addClass(objectClass, aClass)) {
						throw new RuntimeException(
								"A provided instance is not a valid choice ("
										+ aClass + " for " + objectClass + ")");
					}

				if (instances.packages().length > 0) {
					for (String packageName : instances.packages()) {
						ArrayList<Class<?>> classes = Introspection.getClasses(
								new Checker() {
									public boolean accepts(Class<?> aClass) {
										addClass(objectClass, aClass);
										return false;
									}
								}, packageName, 0);
					}
				}
			} else {
				LOGGER.debug("Inspecting class " + objectClass + " for field "
						+ field);
				for (Class<?> aClass : objectClass.getDeclaredClasses()) {
					addClass(objectClass, aClass);
				}
			}
		} else {

			for (Field f : choicesClass.value().getDeclaredFields()) {
				Choice choice = f.getAnnotation(Choice.class);
				if (choicesClass.value() == f.getType()) {
					if (choice == null)
						throw new HandlerException(
								"Field %s [of %s] does not have a ClassChooser.Choice annotation",
								f.toString(), choicesClass.value());
					String name = choice.name().equals("") ? f.getName()
							: choice.name();
					map.put(name, new Information(choice));
				}
			}
		}

		// Remove null values (duplicates)
		Iterator<Information> iterator = map.values().iterator();
		while (iterator.hasNext()) {
			if (iterator.next() == null)
				iterator.remove();
		}
	}

	/**
	 * Add a new possible choice (if valid)
	 * 
	 * @param objectClass
	 *            The base class
	 * @param aClass
	 *            The class to add
	 * @return true if this was choice was added
	 */
	private boolean addClass(Class<? extends Object> objectClass,
			Class<?> aClass) {
		LOGGER.debug("Checking class " + aClass + " for " + objectClass + ": "
				+ objectClass.isAssignableFrom(aClass));

		if (objectClass.isAssignableFrom(aClass)) {
			Choice choice = aClass.getAnnotation(Choice.class);

			Information info = new Information(aClass, choice == null ? null
					: choice.group(), choice == null ? null : choice
					.groupPrefix());

			if (choice == null || choice.name().equals("")) {
				LOGGER
						.debug(String
								.format(
										"The class %s is defined within %s but has no @Choice annotation",
										aClass, objectClass));
				String name = aClass.getName();

				map.put(name, info);
				// Also put the short name
				name = name.substring(name.lastIndexOf('.') + 1);
				if (map.containsKey(name))
					map.put(name, null);
				else
					map.put(name, info);

			} else
				map.put(choice.name(), info);

			return true;
		}

		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * bpiwowar.argparser.handlers.GenericObjectsHandler#getValue(bpiwowar.argparser
	 * .ArgParser, java.lang.String)
	 */
	@Override
	protected Object getValue(ArgParser argParser, String value)
			throws ArgParseException {
		try {
			// Get the matching class
			Information v = map.get(value);
			if (v == null) {
				throw new ArgParseException(
						"Value %s is not valid: the possible choices are %s",
						value, map.keySet());
			}

			// Create an instance of this class
			Constructor<?> constructor = null;
			try {
				constructor = v.theClass.getConstructor();
			} catch (NoSuchMethodException e) {

			}

			Object newInstance;
			if (constructor != null)
				newInstance = constructor.newInstance();
			else {
				// Inner type
				try {
					constructor = v.theClass.getConstructor(object.getClass());
				} catch (NoSuchMethodException e) {
					LOGGER.error("Could not find a proper constructor for "
							+ v.theClass + ": "
							+ Arrays.toString(v.theClass.getConstructors()), e);
				}
				newInstance = constructor.newInstance(object);
			}

			// Set the field with the newly constructed object
			field.set(object, newInstance);

			// Add the options for this object
			String groupPrefix = ArgParser.concatenatePrefixes(prefix, localPrefix, v.groupPrefix);
			LOGGER.debug(format("The class %s was chosen (prefix=%s)", v.theClass, groupPrefix));
			argParser.addOptions(newInstance, true, true, groupPrefix, v.group);
			return newInstance;

		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

}
