/**
 * Copyright Benjamin Piwowarski, 2007. All rights reserved. Permission to use,
 * copy, modify and redistribute is granted, provided that this copyright
 * notice is retained and the author is given credit whenever appropriate.
 *
 * This  software is distributed "as is", without any warranty, including 
 * any implied warranty of merchantability or fitness for a particular
 * use. The author assumes no responsibility for, and shall not be liable
 * for, any special, indirect, or consequential damages, or any damages
 * whatsoever, arising out of or in connection with the use of this
 * software.
 */
package bpiwowar.argparser.handlers;

import java.io.File;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Field;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.StringTokenizer;

import bpiwowar.argparser.ArgParser;
import bpiwowar.argparser.ArgParser.TypeCheckerInterface;
import bpiwowar.argparser.GenericHelper;
import bpiwowar.argparser.StringScanException;
import bpiwowar.argparser.StringScanner;

/**
 * Convert a comma separated list into a collection of files
 * 
 */
@HandlerInformation(FileCollectionHandler.TypeChecker.class)
public class FileCollectionHandler extends GenericObjectsHandler {

	static public class TypeChecker implements TypeCheckerInterface {

		/**
		 * Returns true if the type if ? extends Collection<String>
		 * 
		 * @see bpiwowar.argparser.ArgParser.TypeCheckerInterface#isCompatible(java.lang.Class,
		 *      Field)
		 */
		public boolean isCompatible(Type gtype) {
			return GenericHelper.checkGenerics(Collection.class, gtype, File.class);
		}



		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return "Collection<File>";
		}

	}

	/** Used to give a separator different than the comma */
	@Retention(value = RetentionPolicy.RUNTIME)
	@Target(value = { ElementType.FIELD })
	public @interface Options {
		/** Strings separator */
		String separator() default ",";

		/** Set the object to null if the option is empty */
		boolean nullify() default false;
	}

	String separator = ",";

	boolean nullify = false;

	public FileCollectionHandler(Object object, Field field, Class<?> aClass) {
		super(object, field, aClass);

		// Get the separator
		Options annotation = field.getAnnotation(Options.class);
		if (annotation != null) {
			separator = annotation.separator();
			nullify = annotation.nullify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see argparser.BooleanHolderInterface#isValue()
	 */
	@SuppressWarnings("unchecked")
	public String getValue() {
		try {
			Collection<File> collection = (Collection<File>) get();
			if (collection == null)
				return "";
			String r = null;
			for (File s : collection)
				if (r == null)
					r = s.toString();
				else
					r += separator + s;
			return r;
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see yrla.argparser.holders.Holder#scan(yrla.argparser.StringScanner)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public Object process(ArgParser argParser, StringScanner scanner)
			throws StringScanException {
		try {
			// handles gracefully empty
			String value = scanner.atEnd() ? "" : scanner.scanString();

			StringTokenizer tokenizer = new StringTokenizer(value, separator);
			Collection<File> collection = (Collection<File>) get();
			if (!tokenizer.hasMoreTokens() && nullify) {
				addValue(null);
			} else {
				if (collection == null)
					collection = (Collection<File>) newInstance();
				collection.clear();
				while (tokenizer.hasMoreTokens())
					collection.add(new File(tokenizer.nextToken()));
			}
			return value;
		} catch (Exception e) {
			throw new StringScanException(e);
		}
	}

}
