/**
 * 
 */
package bpiwowar.argparser.handlers;

import java.util.ArrayList;

import org.apache.log4j.Logger;

import bpiwowar.argparser.ArgParseException;
import bpiwowar.argparser.ArgParser;
import bpiwowar.argparser.ArgParserException;
import bpiwowar.argparser.IllegalArgumentValue;
import bpiwowar.argparser.StringScanException;
import bpiwowar.argparser.StringScanner;
import bpiwowar.argparser.checkers.AndCheckers;
import bpiwowar.argparser.checkers.Limit;
import bpiwowar.argparser.checkers.Limit.Type;
import bpiwowar.argparser.checkers.OrChecker;
import bpiwowar.argparser.checkers.ValueChecker;

/**
 * 
 * A generic value handler
 * 
 * @author B. Piwowarski
 * @date Nov 29, 2007
 * 
 */
/**
 * @author bpiwowar
 *
 */
public abstract class Handler {
	final static private Logger LOGGER = Logger.getLogger(Handler.class);

	/**
	 * Our checkers
	 */
	private ArrayList<ValueChecker> checkers = new ArrayList<ValueChecker>();

	/**
	 * Check the value
	 * 
	 * @param o
	 *            The value
	 * @throws IllegalArgumentValue
	 */
	void check(Object value) throws IllegalArgumentValue {
		LOGGER.debug(String.format("We have %d check(s) to do", checkers.size()));
		for (ValueChecker checker : checkers) {
			LOGGER.debug(String.format("Checking with %s", checker));
			boolean check = checker.check(value);
			if (!check) {
				LOGGER.debug("Error in argument " + value);
				throw new IllegalArgumentValue("Checker error: <<" + value
						+ ">> " + checker.getDescription());
			}
		}
	}

	/**
	 * Gets a new limit checker
	 * 
	 * @return a new limit checker
	 * @throws StringScanException
	 * @throws ArgParseException
	 * 
	 */
	public Limit getLimitChecker(StringScanner scanner)
			throws IllegalArgumentException, StringScanException {
		throw new IllegalArgumentException(
				"Bound cannot be defined for holder " + this);
	}

	final public Limit getLimitChecker(Type type, StringScanner scanner)
			throws IllegalArgumentException, StringScanException {
		Limit limitChecker = getLimitChecker(scanner);
		limitChecker.setType(type);
		return limitChecker;
	}

	/**
	 * Is this a help option?
	 * 
	 * @return true if the option is help option
	 */
	public boolean isHelpOption() {
		return false;
	}

	/**
	 * Add a value checker
	 * 
	 * @param index
	 *            The index of the checker
	 * @param newChecker
	 *            the checker to add (or to replace)
	 */

	public void add(ValueChecker newChecker) {
		LOGGER.debug("Adding a new value checker: " + newChecker + " for "
				+ this);
		checkers.add(newChecker);
	}

	/**
	 * Scan a range specification, and add the interval checks to the value
	 * handler
	 * 
	 * @param rec
	 *            the record
	 * @param s
	 *            The string to parse
	 * @throws IllegalArgumentException
	 * @throws StringScanException
	 * @throws
	 * @throws ArgParseException
	 */
	public void updateRangeChecker(OrChecker orChecker, String s)
			throws IllegalArgumentException, StringScanException {
		StringScanner scanner = new StringScanner(s);
		char c;

		scanner.setStringDelimiters(")],}");
		c = scanner.getc(); // swallow the first '{'
		scanner.skipWhiteSpace();

		while ((c = scanner.peekc()) != '}') {

			if (c == '[' || c == '(') {
				AndCheckers rangeChecker = new AndCheckers();

				final char c0 = scanner.getc(); // record & swallow
				// character
				scanner.skipWhiteSpace();

				Limit limitChecker = getLimitChecker(scanner);
				limitChecker.setType(c0 == '(' ? Limit.Type.OPEN_MINIMUM
						: Limit.Type.CLOSED_MINIMUM);
				rangeChecker.add(limitChecker);

				scanner.skipWhiteSpace();
				if (scanner.getc() != ',')
					throw new IllegalArgumentException(
							"Missing ',' in subrange specification");

				limitChecker = getLimitChecker(scanner);
				rangeChecker.add(limitChecker);
				scanner.skipWhiteSpace();
				final char c1 = scanner.getc();
				if (c1 != ']' && c1 != ')')
					throw new IllegalArgumentException("Unterminated subrange");
				limitChecker.setType(c1 == '(' ? Limit.Type.OPEN_MAXIMUM
						: Limit.Type.CLOSED_MAXIMUM);
				scanner.skipWhiteSpace();
				// Add the range checker
				orChecker.add(rangeChecker);
			} else {
				scanner.skipWhiteSpace();
				// Add the equality checker
				orChecker.add(getLimitChecker(Type.EQUALITY, scanner));
			}
			scanner.skipWhiteSpace();
			if ((c = scanner.peekc()) == ',') {
				scanner.getc();
				scanner.skipWhiteSpace();
			} else if (c != '}') {
				throw new IllegalArgumentException(
						"Range spec: ',' or '}' expected (got " + c + ")");
			}
		}
	}

	/**
	 * Get a range checker
	 * 
	 * @param rangeDesc
	 * @return
	 * @throws IllegalArgumentException
	 * @throws StringScanException
	 */
	public OrChecker getRangeChecker(String rangeDesc)
			throws IllegalArgumentException, StringScanException {
		OrChecker orChecker = new OrChecker();
		updateRangeChecker(orChecker, rangeDesc);
		return orChecker;
	}

	/**
	 * Parse the arguments from the current index
	 * 
	 * @param argParser
	 *            the argument parser
	 * @param args
	 *            The set of arguments
	 * @param idx
	 *            The index of the next argument at process
	 * @throws ArgParseException
	 * @throws StringScanException
	 * @throws IllegalArgumentValue
	 *             If the parameter is not OK
	 */
	abstract public int parse(ArgParser argParser, String[] args, int idx)
			throws ArgParserException;


	/**
	 * Called by {@linkplain ArgParser} when the argument had a prefix (to
	 * handle the case of prefix chains)
	 * 
	 * @param prefix
	 *            The prefix
	 */
	public void setPrefix(String prefix) {
	}

	/**
	 * Process the annotations
	 */
	public void processAnnotations() {
	}

}
