/**
 * Copyright Benjamin Piwowarski, 2007. All rights reserved. Permission to use,
 * copy, modify and redistribute is granted, provided that this copyright
 * notice is retained and the author is given credit whenever appropriate.
 *
 * This  software is distributed "as is", without any warranty, including 
 * any implied warranty of merchantability or fitness for a particular
 * use. The author assumes no responsibility for, and shall not be liable
 * for, any special, indirect, or consequential damages, or any damages
 * whatsoever, arising out of or in connection with the use of this
 * software.
 */
package bpiwowar.argparser.handlers;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.StringTokenizer;

import bpiwowar.argparser.ArgParser;
import bpiwowar.argparser.StringScanException;
import bpiwowar.argparser.StringScanner;
import bpiwowar.argparser.ArgParser.TypeCheckerInterface;

/**
 * Convert a comma separated list into an array
 * 
 * @author bpiwowar
 * @date 06/09/2007
 * 
 */
@HandlerInformation(IntegerCollectionHandler.TypeChecker.class)
public class IntegerCollectionHandler extends GenericObjectsHandler {

	static public class TypeChecker implements TypeCheckerInterface {

		/**
		 * Returns true if the type if ? extends Collection<String>
		 * 
		 * @see bpiwowar.argparser.ArgParser.TypeCheckerInterface#isCompatible(java.lang.Class,
		 *      Field)
		 */
		public boolean isCompatible(Type gtype) {
			if (gtype instanceof ParameterizedType) {
				// list the raw type information
				ParameterizedType ptype = (ParameterizedType) gtype;
				Type rtype = ptype.getRawType();
				if (rtype instanceof Class<?>) {
					Class<?> ctype = (Class<?>)rtype;
					if (Collection.class.isAssignableFrom(ctype)) {
						// list the actual type arguments
						Type[] targs = ptype.getActualTypeArguments();
						if (targs.length > 0 && targs[0].equals(Integer.class)) {
							return true;
						}
					}
				}
			}
			
			return false;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return "Collection<String> checker";
		}

	}

	/** Used to give a separator different than the comma */
	@Retention(value = RetentionPolicy.RUNTIME)
	@Target(value = { ElementType.FIELD })
	public @interface Options {
		/** Strings separator */
		String separator() default ",";

		/** Set the object to null if the option is empty */
		boolean nullify() default false;
	}

	String separator = ",";

	boolean nullify = false;

	public IntegerCollectionHandler(Object object, Field field) {
		super(object, field);

		// Get the separator
		Options annotation = field.getAnnotation(Options.class);
		if (annotation != null) {
			separator = annotation.separator();
			nullify = annotation.nullify();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see argparser.BooleanHolderInterface#isValue()
	 */
	@SuppressWarnings("unchecked")
	public String getValue() {
		try {
			Collection<String> collection = (Collection<String>) get();
			if (collection == null)
				return "";
			String r = null;
			for (String s : collection)
				if (r == null)
					r = s;
				else
					r += separator + s;
			return r;
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see yrla.argparser.holders.Holder#scan(yrla.argparser.StringScanner)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public Object process(ArgParser argParser, StringScanner scanner) throws StringScanException {
		try {
			String value = scanner.scanString();
			StringTokenizer tokenizer = new StringTokenizer(value, separator);
			Collection<Integer> collection = (Collection<Integer>) get();
			if (!tokenizer.hasMoreTokens() && nullify) {
				addValue(null);
			} else {
				if (collection == null)
					collection = (Collection<Integer>) newInstance();
				collection.clear();
				while (tokenizer.hasMoreTokens())
					collection.add(Integer.parseInt(tokenizer.nextToken()));
			}
			return value;
		} catch (Exception e) {
			throw new StringScanException(e);
		}
	}

}
