package bpiwowar.argparser.handlers;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import bpiwowar.argparser.ArgParseException;
import bpiwowar.argparser.ArgParser;
import bpiwowar.argparser.ArgParserException;
import bpiwowar.argparser.IllegalArgumentValue;
import bpiwowar.argparser.StringScanException;
import bpiwowar.argparser.holders.ObjectHolder;

/**
 * A handler that calls a method with method arguments extracted from command
 * line
 * 
 * @author B. Piwowarski
 */
public class MethodCallHandler extends Handler {
	final static private Logger logger = Logger
			.getLogger(MethodCallHandler.class);

	private final Object object;
	private final Method method;
	private final Handler[] handlers;
	private final ObjectHolder<?>[] holders;

	private Pattern regexp;

	public MethodCallHandler(Object object, Method method, Handler[] handlers,
			ObjectHolder<?>[] holders, String regexp) {
		this.object = object;
		this.method = method;
		this.handlers = handlers;
		this.holders = holders;
		this.regexp = regexp != null ? Pattern.compile(regexp) : null;
	}

	@Override
	public int parse(ArgParser argParser, String[] args, int idx)
			throws ArgParserException {
		Object[] objects = new Object[handlers.length];

		// --- Use the regular expression if one was given
		if (regexp != null) {
			Matcher matcher = regexp.matcher(args[idx]);
			if (!matcher.matches() || matcher.start() != 0
					|| matcher.end() != args[idx].length())
				throw new IllegalArgumentValue(
						"%s does not match regular expression %s", args[idx],
						regexp);

			String[] args2 = new String[matcher.groupCount()];
			logger.debug("Found " + matcher.groupCount() + " groups in "
					+ args[idx] + " with " + regexp);
			for (int i = 0; i < args2.length; i++) {
				args2[i] = matcher.group(i + 1);
				if (args2[i] == null)
					args2[i] = "";
			}

			logger.debug("Arguments are " + Arrays.toString(args2));
			int idx2 = 0;
			for (int i = 0; i < handlers.length; i++) {
				logger.debug(String.format("Processing arguments with %s",
						handlers[i]));
				idx2 = handlers[i].parse(argParser, args2, idx2);
				objects[i] = holders[i].value;
				logger.debug(String.format("Result is %s", objects[i]));
			}

			if (idx2 != args2.length)
				logger.warn("Some arguments were not read while processing argument with regular expression");

			idx++;
		} else
			for (int i = 0; i < handlers.length; i++) {
				logger.debug(String.format("Processing arguments with %s",
						handlers[i]));
				idx = handlers[i].parse(argParser, args, idx);
				objects[i] = holders[i].value;
				logger.debug(String.format("Result is %s", objects[i]));

			}

		// Call the method
		try {
			boolean accessible = method.isAccessible();
			if (!accessible)
				method.setAccessible(true);
			logger.debug(String.format("Calling the method %s of object %s",
					method, object));
			method.invoke(object, objects);
			if (!accessible)
				method.setAccessible(false);
		} catch (RuntimeException e) {
			throw e;
		} catch (Throwable e) {
			throw new ArgParseException(e,
					"Error (%s) while invoking method %s of object %s", e,
					method, object);
		}
		return idx;
	}
}
