package bpiwowar.experiments;

import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.Constructor;
import java.util.Arrays;

import bpiwowar.argparser.ArgParser;
import bpiwowar.argparser.ArgParserOption;
import bpiwowar.argparser.Logger;

/**
 * An abstract task that takes care of parsing
 * 
 * @author Benjamin Piwowarski <benjamin@bpiwowar.net>
 */
public abstract class AbstractTask implements Task {
	static final bpiwowar.argparser.Logger logger = Logger.getLogger();

	/**
	 * Our parent task
	 */
	Task parent;

	@Retention(RetentionPolicy.RUNTIME)
	public @interface Information {
		/**
		 * The subtasks
		 */
		Class<? extends Task>[] subtasks() default {};

		/**
		 * Synopsis of the command
		 */
		String synopsis() default "[options]";
	}

	/**
	 * Execute the task
	 * 
	 * @param args
	 * @throws Throwable
	 */
	public int execute() throws Throwable {
		return 0;
	}

	@Information
	private static class DefaultAnnotationClass {
	}

	private static final Information DEFAULT_INFORMATION = DefaultAnnotationClass.class
			.getAnnotation(Information.class);

	/**
	 * Get the parent task
	 * 
	 * @return
	 */
	public Task getParent() {
		return parent;
	}

	/**
	 * Process trailing arguments if we have some
	 * 
	 * @param args
	 *            The arguments
	 * @return The unprocessed trailing arguments
	 * @throws Exception
	 */
	public String[] processTrailingArguments(String[] args) throws Exception {
		return args;
	}

	/**
	 * The task to run
	 */
	Task finalTask;

	@Override
	final public void init(String[] args) throws Exception {
		// The task to run is this for now
		logger.info("Initialising task of class %s", getClass()
				.getCanonicalName());
		finalTask = this;

		// Get the information, if any
		Information information = getClass().getAnnotation(Information.class);
		if (information == null)
			information = DEFAULT_INFORMATION;

		String synopsis = String.format("%s %s", getName(),
				information.synopsis());
		if (information.subtasks().length > 0) {
			logger.info("Task has subtasks");
			synopsis += " <task>";
		}

		ArgParser parser = new ArgParser(synopsis);
		parser.addOptions(this, true);

		args = parser.matchAllArgs(args, 0, ArgParserOption.EXIT_ON_ERROR,
				ArgParserOption.STOP_FIRST_UNMATCHED);

		// Process what remains
		args = processTrailingArguments(args);

		if (information.subtasks().length == 0) {
			// Check that we don't have trailing arguments
			if (args != null && args.length > 0)
				throw new RuntimeException(String.format(
						"Unprocessed command line arguments: %s",
						bpiwowar.argparser.utils.Output.toString(" ", args)));
		} else {

			// We have subtasks
			if (args == null || args.length == 0) {
				System.err.println("Possible subtasks are:");
				// Output the possible subtasks
				for (Class<? extends Task> subtask : information.subtasks()) {
					Subtask annotation = subtask.getAnnotation(Subtask.class);
					if (annotation == null)
						logger.error("Class %s has no Subtask annotation",
								subtask.getCanonicalName());
					else
						System.err.format("%s\t%s%n", annotation.name(),
								annotation.description());
				}
				return;
			}

			// --- Try to match the subclass
			Class<? extends Task> subtaskClass = null;
			String subtaskName = args[0];
			for (Class<? extends Task> aSubtask : information.subtasks()) {
				Subtask info = aSubtask.getAnnotation(Subtask.class);
				if (info == null)
					logger.error("Class %s has no Subtask annotation",
							aSubtask.getCanonicalName());

				else if (info.name().equals(subtaskName)) {
					if (subtaskClass != null)
						throw new RuntimeException(String.format(
								"Task name %s has multiple matches",
								subtaskName));
					subtaskClass = aSubtask;
				}
			}

			if (subtaskClass == null)
				throw new RuntimeException("Task " + subtaskName
						+ " could not be found");

			// --- Run the subtask
			Subtask info = subtaskClass.getAnnotation(Subtask.class);
			logger.info("Initialising sub-task %s [%s]", info.name(),
					subtaskClass.getCanonicalName());

			Task subtask;
			if (subtaskClass.isMemberClass()) {
				// If it is a class enclosed in the main task class, then
				// we construct it with us
				Constructor<? extends Task> constructor = subtaskClass
						.getDeclaredConstructor(getClass());
				subtask = constructor.newInstance(this);
			} else {
				subtask = subtaskClass.newInstance();
				if (subtask instanceof AbstractTask)
					((AbstractTask) subtask).parent = this;
			}

			// Initialise with remaining arguments
			args = Arrays.copyOfRange(args, 1, args.length);
			subtask.init(args);
			finalTask = subtask;
		}
	}

	final public int run() throws Throwable {
		logger.info("Running task of class %s", finalTask.getClass());
		if (finalTask instanceof AbstractTask)
			return ((AbstractTask) finalTask).execute();

		return finalTask.run();
	}

	/**
	 * Returns the name of the task
	 * 
	 * @return The experiment name annotation, or the class name if there is no
	 *         such annotation
	 */
	protected String getName() {
		TaskDescription annotation = getClass().getAnnotation(
				TaskDescription.class);
		if (annotation != null)
			return annotation.name();
		return getClass().getCanonicalName();
	}

}
